/* FILE: eulerevolve.h                 -*-Mode: c++-*-
 *
 * Concrete evolver class, using simple forward Euler steps
 *
 */

#ifndef _OXS_EULEREVOLVE
#define _OXS_EULEREVOLVE

#include "driver.h"
#include "evolver.h"
#include "key.h"

/* End includes */

class Oxs_EulerEvolve:public Oxs_Evolver {
private:
  // Step size control parameters.  Each may be disabled by setting
  // to -1.  There is an additional step size control that insures
  // that energy is monotonically non-increasing (up to estimated
  // rounding error).
  REAL8m allowed_error_rate;  // Step size is adjusted so
  /// that the estimated maximum error (across all spins) divided
  /// by the step size is smaller than this value.  The units
  /// internally are radians per second, converted from the value
  /// specified in the input MIF file, which is in deg/sec.

  REAL8m allowed_absolute_step_error; // Similar to allowed_error_rate,
  /// but without the step size adjustment.  Internal units are
  /// radians; MIF input units are degrees.

  REAL8m allowed_relative_step_error; // Step size is adjusted so that
  /// the estimated maximum error (across all spins) divided by
  /// [maximum dm/dt (across all spins) * step size] is smaller than
  /// this value.  This value is non-dimensional, representing the
  /// allowed relative (proportional) error, presumably in (0,1).

  REAL8m step_headroom; // The 3 control parameters above can be
  /// used to estimate the step size that would just fit the control
  /// requirements.  Because this is only an estimate, if the step size
  /// is actually set to that value there is a good chance that the
  /// requirement will not be met.  So instead, we leave some headroom
  /// by setting the step size to the computed value multiplied by
  /// step_headroom.  This is a non-dimensional quantity, which should
  /// be in the range (0,1).

  // The total energy field in Oxs_SimState is computed by accumulating
  // the dE into the total energy from the previous state.  It order to
  // keep this value from becoming too inaccurate, we recalculate total
  // energy directly from the energy densities after each
  // energy_accum_count_limit passes.  NB: The code here assumes that a
  // single sequence of states is being fed to this routine.  If this
  // is not the case, then the accum count needs to be tied to the state
  // id.
  const UINT4m energy_accum_count_limit ;
  UINT4m energy_accum_count;

  // The following evolution constants are uniform for now.  These
  // should be changed to arrays in the future.
  REAL8m gamma;  // Landau-Lifschitz gyromagnetic ratio
  REAL8m alpha;  // Landau-Lifschitz damping coef
  BOOL do_precess;  // If false, then do pure damping

  // next_timestep is based on the error from the last step.  If
  // there is no last step (either because this is the first step,
  // or because the last state handled by this routine is different
  // from the incoming current_state), then timestep is calculated
  // so that max_dm_dt * timestep = start_dm.
  REAL8m start_dm;
  REAL8m next_timestep;

  // Data cached from last state
  UINT4m state_id;
  REAL8m max_dm_dt;
  REAL8m timestep_lower_bound;  // Smallest timestep that can actually
  /// change spin with max_dm_dt (due to REAL8_EPSILON restrictions).
  REAL8m dE_dt;
  REAL8m total_energy;
  REAL8m total_energy_change;
  Oxs_MeshValue<ThreeVector> dm_dt;
  Oxs_MeshValue<REAL8m> energy;

  // Outputs
  void UpdateDerivedOutputs(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_EulerEvolve> max_dm_dt_output;
  Oxs_ScalarOutput<Oxs_EulerEvolve> dE_dt_output;
  Oxs_ScalarOutput<Oxs_EulerEvolve> total_energy_output;
  Oxs_VectorFieldOutput<Oxs_EulerEvolve> dm_dt_output;
  Oxs_VectorFieldOutput<Oxs_EulerEvolve> total_H_field_output;

  // Scratch space
  Oxs_MeshValue<REAL8m> new_energy;
  Oxs_MeshValue<ThreeVector> new_dm_dt;
  Oxs_MeshValue<ThreeVector> mxH;
  Oxs_MeshValue<ThreeVector> new_H;

  void Calculate_dm_dt
  (const Oxs_Mesh& mesh_,
   const Oxs_MeshValue<REAL8m>& Ms_,
   const Oxs_MeshValue<ThreeVector>& mxH_,
   const Oxs_MeshValue<ThreeVector>& spin_,
   Oxs_MeshValue<ThreeVector>& dm_dt_,
   REAL8m& max_dm_dt_,REAL8m& dE_dt_,REAL8m& min_timestep_) const;
  /// Imports: mesh_, Ms_, mxH_, spin_
  /// Exports: dm_dt_, max_dm_dt_, max_dm_dt_index, dE_dt_

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  virtual BOOL Init();
  Oxs_EulerEvolve(const char* name,     // Child instance id
		 Oxs_Director* newdtr, // App director
		 Tcl_Interp* safe_interp, // Safe interpreter
		 const char* argstr);  // MIF input block parameters
  virtual ~Oxs_EulerEvolve();

  virtual  BOOL
  Step(const Oxs_Driver& driver,
       Oxs_ConstKey<Oxs_SimState> current_state,
       Oxs_Key<Oxs_SimState>& next_state);
  // Returns true if step was successful, false if
  // unable to step as requested.
};

#endif // _OXS_EULEREVOLVE
