/* FILE: vecfile.h                 -*-Mode: c++-*-
 *
 * Header file for vector field file I/O.
 *
 */

#ifndef _VF_VECFILE
#define _VF_VECFILE

#include <stdio.h>

#include "nb.h"

#include "mesh.h"

class Vf_FileInput;  // Forward declaration

//////////////////////////////////////////////////////////////////////
// Vf_FileInputID
//
class Vf_FileInputID
{
  friend class Vf_FileInput;
private:
  static const ClassDoc class_doc;
  Nb_DString filetype;
  BOOL (*pIsType)(const char *filename,
		  FILE *input_file,
		  Tcl_Channel channel,
		  const char *channelName);
  Vf_FileInput* (*pMakeNew)(const char *new_filename,
                            FILE *new_infile,
                            Tcl_Channel new_channel,
                            const char *new_channelName);
  void Dup(const Vf_FileInputID &rhs); // Copy
public:
  // Constructor *with* registration (for global initialization)
  // (Set last field to NULL to avoid registration.)
  Vf_FileInputID(const char *new_filetype,
                 BOOL (*newpIsType)(const char *filename,
                                    FILE *input_file,
                                    Tcl_Channel channel,
                                    const char *channelName),
                 Vf_FileInput *
                 (*newpNewReader)(const char *filename,
                                  FILE *input_file,
                                  Tcl_Channel channel,
                                  const char *channelName),
                 BOOL (*pRegister)(const Vf_FileInputID &)
                 );

  // Copy constructor
  Vf_FileInputID(const Vf_FileInputID &copyobj);

  Vf_FileInputID(); // Default constructor; Needed by Nb_List<T>
  Vf_FileInputID& operator=(const Vf_FileInputID &rhs);

  ~Vf_FileInputID();
};

//////////////////////////////////////////////////////////////////////
// Vf_FileInput
//
class Vf_FileInput
{
  // NOTES ON CHILD CLASSES:
  //
  // Every child class should have a static member function defined
  // like this:
  //
  // static Vf_FileInput* MakeNew(const char *new_filename,
  //                              FILE *new_infile,
  //                              Tcl_Channel new_channel,
  //	                          const char *new_channelName);
  //
  // which calls its "internal" constructor.  This function is passed
  // to Vf_FileInput::RegisterChild as part of a Vf_FileInputID.
  //
  // All children classes should have a static member variable:
  //
  //   static Vf_FileInputID id;
  //
  // This initialization of this variable should call
  // Vf_FileInput::RegisterChild(id) so that they can be
  // called as necessary from Vf_FileInput::NewReader().
  //
  // The FILE *infile is closed in the base class destructor.

private:
  static const ClassDoc class_doc;
  static Nb_List<Vf_FileInputID> child_list;

  // The following 3 functions are left undefined on purpose
  Vf_FileInput(); // Default constructor
  Vf_FileInput(Vf_FileInput &copyobj);  // Copy constructor
  Vf_FileInput& operator=(Vf_FileInput &rhs); // Assignment op

protected:
  const Nb_DString filename;
  Nb_DString filetitle; // (Long) filetitle
  Nb_DString filedesc;  // File description
  FILE *infile;
  Tcl_Channel channel;     // Channel registered with omfInterp
  const char *channelName; // (omfInterp is a global variable defined
  // in tclcmds.cc, which is a pointer to the current Tcl command
  // interpreter.)
  // NOTE: The file is "registered" with the Tcl interpreter inside
  // "NewReader", and so must be closed *only* by a corresponding
  // call to Tcl_UnregisterChannel(), which occurs in the
  // ~Vf_FileInput destructor, NOT IN ANY OF THE CHILD CLASSES!

  // Internal constructor
  Vf_FileInput(const char *new_filename,
		       FILE *new_infile,
		       Tcl_Channel new_channel,
		       const char *new_channelName);

  static BOOL RegisterChild(const Vf_FileInputID &id);

public:
  // Clients should call NewReader() to get a VFFI instance.  The
  // client is responsible for destroying (via delete) the returned
  // object, i.e., treat this as a "new" operator.
  static Vf_FileInput *NewReader(const char *new_filename);

  virtual ~Vf_FileInput();
  const char *GetFileName(void) { return (const char *)filename; }
  virtual const char *FileType(void)=0;
  virtual Vf_Mesh *NewMesh(void)=0;  // Derived classes should return
  /// a *Vf_EmptyMesh on error (not NULL).  Either way, the client
  /// is responsible for destroying (via delete) the returned mesh.
};

//////////////////////////////////////////////////////////////////////
// Omf File Format specs and utility routines
enum Vf_OvfDataStyle { vf_oascii, vf_obin4, vf_obin8 };
class Vf_OvfFileFormatSpecs
{
private:
  static const ClassDoc class_doc;
  static const REAL4 CheckValue4;  // 4-byte real data block check value
  static const REAL8 CheckValue8;  // 8-byte   "                     "
public:
  // Format of data in "Data" block: ASCII (text),
  // 4-byte IEEE binary, 8-byte IEEE binary

  static BOOL CheckIDString(const char *test); // Checks "Magic" string
  /// at top of .ovf files.
  static BOOL CheckValue(REAL4 test) { return test==CheckValue4; }
  static BOOL CheckValue(REAL8 test) { return test==CheckValue8; }

  // The use of the CheckValue functions is encouraged, but
  // the actual values are needed for creating ovf files.
  static void GetCheckValue(REAL4 &check) { check=CheckValue4; }
  static void GetCheckValue(REAL8 &check) { check=CheckValue8; }

  // Binary I/O routines that perform byteswapping as needed.
  static INT4m ReadBinary(FILE *infile,REAL4 *buf,UINT4m count);
  static INT4m ReadBinary(FILE *infile,REAL8 *buf,UINT4m count);
  static INT4m WriteBinary(FILE *outfile,REAL4 *buf,UINT4m count);
  static INT4m WriteBinary(FILE *outfile,REAL8 *buf,UINT4m count);
  static INT4m WriteBinary(Tcl_Channel channel,REAL4 *buf,UINT4m count);
  static INT4m WriteBinary(Tcl_Channel channel,REAL8 *buf,UINT4m count);
  // NOTE: The write routines may modify buf during byteswapping.
  //   If this is unacceptable, then one should add additional
  //   routines that take a const buf, and copy buf into internal
  //   storage.

};

//////////////////////////////////////////////////////////////////////
// Vf_OvfSegmentHeader
//
enum Vf_OvfSegmentHeadMeshType { vf_oshmt_rectangular, vf_oshmt_irregular  };
struct Vf_OvfSegmentHeader
{
private:
  static const ClassDoc class_doc;
public:
  // HEADER VALUES
  Nb_DString title; // Long file name or title goes here
  Nb_DString desc;  // File description
  Nb_DString meshunit; // Fundamental mesh measurement unit. 
                      /// Treated as a label. Example: nm
  Vf_OvfSegmentHeadMeshType meshtype;

  REAL8 xbase,ybase,zbase; // (xbase,ybase,zbase) is the position, in
                           // 'meshunit', of the first point in the data
                           //  section (below).

  REAL8 xstepsize,ystepsize,zstepsize;  // Distance between adjacent
  // grid pts.: on the x-axis, 20 nm, etc.  The sign on these values
  // determines the grid orientation relative to (xbase,ybase,zbase).

  INT4 xnodes,ynodes,znodes; // Number of nodes along each axis, regular grids.
  INT4 pointcount;           // Number of points, for irregular grids.

  REAL8 xmin,ymin,zmin;  // Corner points defining mesh boundary.  The
  REAL8 xmax,ymax,zmax;  // bounding box is set to include both these
  /// points and the extent from ?base, ?stepsize and ?nodes.

  Nb_List< Nb_Vec3<REAL8> > bdryline;  // Points defining a polyline, used to
  /// indicate boundary.  Optional; if missing, a bdry will be constructed
  /// from the ?min & ?max boundary points.

  Nb_DString valueunit; // Fundamental field value unit, treated as a
                       /// label. Example: kA/m

  REAL8 valuemultiplier; // Multiply file values by this to get
                         // true value in 'valueunits'.

  REAL8 valuerangemaxmag,valuerangeminmag; // These are in file value
  // units, and are used as hints (or defaults) by postprocessing programs.
  //  The mmdisp program ignores any points with magnitude smaller than
  // valuerangeminmag, and uses valuerangemaxmag to scale inputs for display.
  // The use of these values is postprocessor dependent, and should be
  // interactively overrideable, but typically ValueRangeMaxMag should be
  // the maximum vector magnitude in the file data, and ValueRangeMinMag
  // should be a lower bound on non-zero file vector magnitudes.

  // BOOLEANS
  // The following boolean values indicate whether each of the above values
  // have been set (1=>set, 0=>not set).
  BOOL title_bool,desc_bool;
  BOOL meshunit_bool,xbase_bool,ybase_bool,zbase_bool;
  BOOL xstepsize_bool,ystepsize_bool,zstepsize_bool;
  BOOL xnodes_bool,ynodes_bool,znodes_bool;
  BOOL xmin_bool,ymin_bool,zmin_bool;
  BOOL xmax_bool,ymax_bool,zmax_bool;
  BOOL valueunit_bool,valuemultiplier_bool;
  BOOL valuerangemaxmag_bool,valuerangeminmag_bool;
  BOOL meshtype_bool,pointcount_bool,bdryline_bool;

  // Utility routines
  Vf_OvfSegmentHeader();
  ~Vf_OvfSegmentHeader();
  void SetBools(BOOL newval);  // Sets all BOOL's to newval
  BOOL CompleteSet(); // Returns 1 iff a complete collection of
  /// the BOOL's are 1.
  void SetFieldValue(const Nb_DString &field,const Nb_DString &value);
};

//////////////////////////////////////////////////////////////////////
// Vf_OvfFileInput
//
class Vf_OvfFileInput : public Vf_FileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const Vf_FileInputID id;

  // File read helper routines
  INT4m ReadLine(Nb_DString &buf);
  INT4m ParseContentLine(const Nb_DString &buf,
			 Nb_DString &field,Nb_DString &value);
  INT4m ReadTextVec3f(Nb_Vec3<REAL8> &data);
  INT4m ReadVec3f(Vf_OvfDataStyle datastyle,
		  Nb_Vec3<REAL8> &data,
		  REAL8m* badvalue=NULL);

  // Internal constructor
  Vf_OvfFileInput(const char *new_filename,
	       FILE *new_infile,
	       Tcl_Channel new_channel,
	       const char *new_channelName);

  // The next 2 functions & the following "Default constructor"
  // are left undefined on purpose
  Vf_OvfFileInput(Vf_OvfFileInput &copyobj);  // Copy constructor
  Vf_OvfFileInput& operator=(Vf_OvfFileInput &rhs); // Assignment op

public:
  Vf_OvfFileInput();  // Default constructor
  // Once upon a time (sometime before Aug-1997), I had trouble
  // with the SGI CC compiler requiring at least 1 public
  // constructor, so I put the dummy "Vf_OvfFileInput();" constructor
  // declaration here (but with no definition).  This restriction
  // appears to have gone away with the newer IRIX 6.2 CC compiler,
  // but I'll leave this note here just in case.  Of course, actual
  // code is required to call "MakeNew" instead of a constructor.

  static Vf_FileInput* MakeNew(const char *new_filename,
				       FILE *new_infile,
				       Tcl_Channel new_channel,
				       const char *new_channelName);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(const char *filename,
		     FILE *input_file,
		     Tcl_Channel channel,
		     const char *channelName);
  Vf_Mesh *NewMesh(void);
};

//////////////////////////////////////////////////////////////////////
// Vf_OvfFileOutput
//
class Vf_OvfFileOutput
{
private:
  static const ClassDoc class_doc;
public:
  Vf_OvfFileOutput() {}
  INT4m WriteMesh(Vf_Mesh *mesh,FILE* fptr,Vf_OvfDataStyle datastyle,
		  BOOL force_irreg=0,
		  const char *title=NULL,const char *desc=NULL);
  // Returns -1 if mesh type not supported,
  //          0 on success,
  //         >0 on other error.
  // If force_irreg is set, the output will be in the Omf irregular
  // grid file format (location & value) even if input data lies
  // on a rectangular grid.
  ~Vf_OvfFileOutput() {}
};

//////////////////////////////////////////////////////////////////////
// Vf_VioFileOutput
//
class Vf_VioFileOutput
{
private:
  static const ClassDoc class_doc;
public:
  Vf_VioFileOutput() {}
  INT4m WriteMesh(Vf_Mesh *mesh,FILE* fptr,const char *desc=NULL);
  // Returns -1 if mesh type not supported,
  //          0 on success,
  //         >0 on other error.
  ~Vf_VioFileOutput() {}
};

//////////////////////////////////////////////////////////////////////
// Vf_VioFileInput
//
class Vf_VioFileInput : public Vf_FileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const Vf_FileInputID id;

  // The following 3 functions are left undefined on purpose
  Vf_VioFileInput(); // Default constructor
  Vf_VioFileInput(Vf_OvfFileInput &copyobj);  // Copy constructor
  Vf_VioFileInput& operator=(Vf_OvfFileInput &rhs); // Assignment op

public:
  // Internal constructor
  Vf_VioFileInput(const char *new_filename,
	       FILE *new_infile,
	       Tcl_Channel new_channel,
	       const char *new_channelName);

  // Standard constructor
  static Vf_FileInput* MakeNew(const char *new_filename,
				       FILE *new_infile,
				       Tcl_Channel new_channel,
				       const char *new_channelName);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(const char *filename,
		     FILE *input_file,
		     Tcl_Channel channel,
		     const char *channelName);
  Vf_Mesh *NewMesh(void); // Returns *Vf_EmptyMesh on error
  Vf_GridVec3f *NewRectGrid(void);  // Return NULL on error
};

//////////////////////////////////////////////////////////////////////
// Standard problem (or Simple) Vector Format (SVF)
//

Tcl_CmdProc Vf_SvfFileInputTA;

class Vf_SvfFileInput : public Vf_FileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const Vf_FileInputID id;
  friend int Vf_SvfFileInputTA(ClientData,Tcl_Interp *interp,
                               int argc,char** argv);

  // Instance data members
  Vf_GeneralMesh3f *mesh;

  // The following 3 functions are left undefined on purpose
  Vf_SvfFileInput(); // Default constructor
  Vf_SvfFileInput(Vf_OvfFileInput &copyobj);  // Copy constructor
  Vf_SvfFileInput& operator=(Vf_OvfFileInput &rhs); // Assignment op

  // Internal constructor
  Vf_SvfFileInput(const char *new_filename,
	       FILE *new_infile,
	       Tcl_Channel new_channel,
	       const char *new_channelName);

public:

  // Standard constructor
  static Vf_FileInput* MakeNew(const char *new_filename,
				       FILE *new_infile,
				       Tcl_Channel new_channel,
				       const char *new_channelName);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(const char *filename,
		     FILE *input_file,
		     Tcl_Channel channel,
		     const char *channelName);
  Vf_Mesh *NewMesh(void); // Returns *Vf_EmptyMesh on error
  Vf_GridVec3f *NewRectGrid(void) { return NULL; } // Return NULL on error
  ~Vf_SvfFileInput();
};

#endif // _VF_VECFILE
